
"""
# -- --------------------------------------------------------------------------------------------------- -- #
# -- Project: lucidmode                                                                                  -- #
# -- Description: A Lightweight Framework with Transparent and Interpretable Machine Learning Models     -- #
# -- functions.py: python script with math functions                                                     -- #
# -- Author: IFFranciscoME - if.francisco.me@gmail.com                                                   -- #
# -- license: GPL-3.0 License                                                                            -- #
# -- Repository: https://github.com/lucidmode/lucidmode                                                  -- #
# -- --------------------------------------------------------------------------------------------------- -- #
"""

# -- Load libraries for script
import numpy as np

# --------------------------------------------------------------------------------------- COST FUNCTIONS -- #
# --------------------------------------------------------------------------------------------------------- #

def _cost(A, Y, type):
    
    # numerical stability parameter
    ns = 1e-25
    A = A + ns

    # -- Mean Squared Error
    if type == 'sse':
        
        # loss as the difference on prediction
        loss = A - Y
        # cost as the sum of the squared errors
        cost = np.sum(((loss)**2))

    # -- Binary Cross-Entropy (pending)
    elif type == 'binary-logloss':
        
        # loss as the errors within each value
        loss = np.multiply(Y, np.log(A)) + np.multiply(1 - Y, np.log(1 - A))
        # cost as the mean of loss
        cost = -(1/Y.shape[1]) * np.sum(loss)
    
    # -- Multiclass Cross-Entropy (pending)
    elif type == 'multi-logloss':

        # auxiliary object
        y_hat = np.zeros(shape=(Y.shape[0], A.shape[1]))
        y_hat[range(len(Y)), Y] = 1

        # loss as the errors within each value
        loss = np.sum(-y_hat * np.log(A))
        # cost as the mean of loss
        cost = np.sum(loss)/y_hat.shape[0]
    
    # compress to 0 dimension (to have a float)
    cost = np.squeeze(cost)

    # check final dimensions
    assert(cost.shape == ())

    # function final result
    return cost.astype(np.float32).round(decimals=4)
 
# --------------------------------------------------------------------------------- ACTIVATION FUNCTIONS -- #
# --------------------------------------------------------------------------------------------------------- #

def _sigma(Z, activation):

    # -- Sigmoidal (sigmoid)
    if activation == 'sigmoid':
        Z = 1 / (1 + np.exp(-Z))
        Z = Z.astype(np.float32)
        return Z
    
    # -- Hyperbolic Tangent (tanh)
    elif activation == 'tanh':
        Z = np.tanh(Z)
        Z = Z.astype(np.float32)
        return Z

    # -- Rectified Linear Unit (ReLU)
    elif activation == 'relu':
        A = np.maximum(0, Z)
        assert(A.shape == Z.shape)
        return A
    
    # -- Softmax
    elif activation == 'softmax':

        expZ = np.exp(Z - np.max(Z)).T 
        Z = (expZ / expZ.sum(axis=0, keepdims=True)).T 
        Z = Z.astype(np.float32)
        return Z

# ------------------------------------------------------------------- DERIVATIVE OF ACTIVATION FUNCTIONS -- #
# --------------------------------------------------------------------------------------------------------- #

def _dsigma(Z, activation):
    
    # -- Sigmoid
    if activation == 'sigmoid':
        s = _sigma(Z, activation)
        dZ = s*(1-s)
        dZ = dZ.astype(np.float32)
        assert (dZ.shape == Z.shape)
    
    # -- Hyperbolic Tangent
    elif activation == 'tanh':
        a = _sigma(Z, activation)
        dZ = 1 - a**2
        dZ = dZ.astype(np.float32)
        assert (dZ.shape == Z.shape)
    
    # -- Rectified Linear Unit (ReLU)
    elif activation == 'relu':
        dZ = np.array(Z, copy=True)
        dZ[Z <= 0] = 0
        dZ[Z > 0] = 1
        dZ = dZ.astype(np.int8)
        assert (dZ.shape == Z.shape)
    
    # -- Softmax
    elif activation == 'softmax':
        s = _sigma(Z, activation)
        s = s.reshape(-1, 1)
        s = s.astype(np.float32)
        dZ = np.diagflat(s) - np.dot(s, s.T)
        assert (dZ.shape == Z.shape)
    
    return dZ




"""
# -- --------------------------------------------------------------------------------------------------- -- #
# -- Project: lucidmode                                                                                  -- #
# -- Description: A Lightweight Framework with Transparent and Interpretable Machine Learning Models     -- #
# -- visualizations.py: python script with visualization functions                                       -- #
# -- Author: IFFranciscoME - if.francisco.me@gmail.com                                                   -- #
# -- license: GPL-3.0 License                                                                            -- #
# -- Repository: https://github.com/lucidmode/lucidmode                                                  -- #
# -- --------------------------------------------------------------------------------------------------- -- #
"""

# -- Load libraries for script
import numpy as np

# ---------------------------------------------------------------------- L1-L2-ELASTICNET REGULARIZATION -- #
# --------------------------------------------------------------------------------------------------------- #

def _l1_l2_EN(n_weights, type=None, lmbda=0, ratio=0):
    """
    Regularization operation, can perform 3 types, L1 (Lasso), L2 (Ridge), Elastic Net = ratio(L1 - L2)

    In Cost Function:
    - Weights values of all layers (l1, l2, Elastic Net)

    In weights updating process of selected layers:
    - Applied to weights values (l1, l2, Elastic Net)
    - Applied to bias values (l1, l2, Elastic Net)

    Parameters
    ----------

    n_weights: list of np.array
        
        Containing the weights values to be regularized

    type: {'l1', 'l2', 'elasticnet'}, default='l1' 
        
        'l1': Lasso --> (lambda) * ||weights||_1
        'l2': Ridge --> (lambda) * 0.5 * ||weights||^2_2
        'elasticnet': ratio * L1 + (1 - ratio) * L2

    lmbda: float
        
        value for lambda term in the regularization, between 0 and 1

    ratio: float
        
        Used with the option type='elasticnet', is the ratio between l1 and l2 impact on the regularization

    Return
    ------
    
    reg_value: float
        
        The result of the calculation squeezed to shape (1,)


    References
    ----------

    [3] Zou and Hastie. Regularization and variable selection via the elastic net. Journal Of The Royal     
        Statistical Society 67(2):301 - 320, 2005.

    """

    if type == None:
        reg = 0

    elif type == 'l1':
        reg = 0
        for weights in n_weights:
            reg += (lmbda) * np.sum(abs(weights))

    elif type == 'l2':
        reg = 0
        for weights in n_weights:
            reg += (lmbda) * 0.5 * np.sum(weights**2)

    elif type == 'elasticnet':
        reg = 0
        for weights in n_weights:
            reg += ratio*((lmbda) * np.sum(abs(weights))) + (1 - ratio)*((lmbda) * 0.5 * np.sum(weights**2))
    
    return reg.astype(np.float32)

# ------------------------------------------------------------------------------- DROPOUT REGULARIZATION -- #
# --------------------------------------------------------------------------------------------------------- #

# - In layers
# - Neurons activation (dropout)



"""Transparent and Interpretable Machine Learning Models in Python, with a keras inspired API

``lucidmode`` is a set of interpretable Machine Learning Models.

"""
__version__ = '0.4.3.1'

__all__ = ['functions', 'models', 'propagate', 'regularization']




"""
# -- --------------------------------------------------------------------------------------------------- -- #
# -- Project: lucidmode                                                                                  -- #
# -- Description: A Lightweight Framework with Transparent and Interpretable Machine Learning Models     -- #
# -- propagate.py: python script with forward and backward propagation functions                         -- #
# -- Author: IFFranciscoME - if.francisco.me@gmail.com                                                   -- #
# -- license: GPL-3.0 License                                                                            -- #
# -- Repository: https://github.com/lucidmode/lucidmode                                                  -- #
# -- --------------------------------------------------------------------------------------------------- -- #
"""

# -- Load libraries for script
import numpy as np

# -- Load other scripts
import lucidmode.functions as fn
import tools.metrics as mt
import lucidmode.regularization as rg

# ---------------------------------------------------------------------------------------------- FORWARD -- #

def __forward(self, A, l):
    layer = list(self.layers.keys())[l]
    W = self.layers[layer]['W']
    b = self.layers[layer]['b']
    f = np.matmul(A, W.T) + b.T
    
    return f.astype(np.float16)

def __forward_activation(self, A_prev, l):
    layer = list(self.layers.keys())[l]
    Z = __forward(self, A_prev, l)
    A = fn._sigma(Z, self.layers[layer]['a'])

    return A, Z

def _forward_propagate(self, X): 
    
    # memory to store all the values for later use in backward process
    memory = {'A_' + str(i): 0 for i in range(1, len(self.hidden_l) + 3)}
    memory.update({'Z_' + str(i): 0 for i in range(1, len(self.hidden_l) + 2)})
    memory.update({'d_' + str(i): 0 for i in range(2, len(self.hidden_l) + 3)})
    memory.update({'dW_' + str(i): 0 for i in range(1, len(self.hidden_l) + 2)})
    memory.update({'db_' + str(i): 0 for i in range(1, len(self.hidden_l) + 2)})
    Al, memory['A_1'] = X, X

    for l in range(0, len(self.hidden_l) + 1):
        A_prev = Al
        Al, Zl = __forward_activation(self, A_prev, l)
        # save A and Z for every layer (for backward process)
        memory['Z_' + str(l + 1)] = Zl
        memory['A_' + str(l + 2)] = Al

    return memory


# --------------------------------------------------------------------------------------------- BACKWARD -- #


def __backward_propagate(self, memory, Y):
    
    # number of samples to scale gradients values
    m = memory['A_1'].shape[0]
    
    # -- MULTI-CLASS: perform one-hot, simple difference for output delta
    if self.output_n != 1:
        # one-hot encoding
        one_hot = np.zeros(shape=(len(Y), self.output_n))
        one_hot[range(len(Y)), Y] = 1
        Y = one_hot.astype(np.int8)
        one_hot = None
        # get the post-activation values for the last layer
        AL = memory['A_' + str(len(self.hidden_l) + 2)]
        # first delta for output layer
        dAL = (AL - Y)

    # -- SINGLE-CLASS: Plain variable usage, activated difference for output delta
    else: 
        # get the post-activation values for the last layer
        AL = memory['A_' + str(len(self.hidden_l) + 2)] + 1e-20 # to avoid 0 division 
        Y = Y.reshape(AL.shape)   
        # first delta for output layer
        dAL = (AL - Y)*fn._dsigma(memory['Z_' + str(len(self.hidden_l) + 1)], self.output_a)

    # store output layer delta
    memory['d_' + str(len(self.hidden_l) + 2)] = dAL

    # just loop hidden layers since the above was for the outputlayer
    for l in range(len(self.hidden_l) - 1 , -1, -1):

        # layer labels
        layer = list(self.layers.keys())[l]
        layer_1 = list(self.layers.keys())[l+1]

        # dW and db previous layer
        dW = (1/m) * np.dot(memory['d_' + str(l + 3)].T, memory['A_' + str(l + 2)])
        memory['dW_' + str(l + 2)] = dW
        db = (1/m) * np.sum(memory['d_' + str(l + 3)], axis=0).reshape(dW.shape[0], 1)
        memory['db_' + str(l + 2)] = db
        
        # delta of layer
        delta = fn._dsigma(memory['A_' + str(l + 2)], self.layers[layer]['a'])
        d = delta * np.matmul(memory['d_' + str(l + 3)], self.layers[layer_1]['W'])
        memory['d_' + str(l + 2)] = d

        # check for dimensions
        assert (d.shape == memory['A_' + str(l + 2)].shape)
        assert (dW.shape == self.layers[layer_1]['W'].shape)
        assert (db.shape == self.layers[layer_1]['b'].shape)           

    # last delta for the input layer
    memory['dW_1'] =  (1/m) * np.dot(memory['d_2'].T, memory['A_1'])
    memory['db_1'] =  (1/m) * sum(memory['d_2']).reshape(self.layers['hl_0']['W'].shape[0], 1)

    return memory


# --------------------------------------------------------------------------------- FORWARD <-> BACKWARD -- #


def _forward_backward(self, x_train, y_train, x_val=None, y_val=None, epoch=0, verbosity=3, metric_goal=None):
    """
    """

    # -- Calculations for Train set

    # Forward pass
    memory_train = _forward_propagate(self, x_train)
    mem_layer = 'A_' + str(len(self.hidden_l) + 2)

    # Probability prediction and cost value calculation for train
    y_train_p = memory_train[mem_layer]
    y_train_hat = self.predict(x_train)
    cost_train = fn._cost(y_train_p, y_train, self.cost['function'])

    # -- Calculations for Validation set
    if len(x_val) !=0:

        # Forward pass, cost and prediction value calculations for val
        memory_val = _forward_propagate(self, x_val)
        y_val_hat = memory_val[mem_layer]
        cost_val = fn._cost(y_val_hat, y_val, self.cost['function'])
        self.history[self.cost['function']]['val'][epoch] = cost_val
        y_val_hat = self.predict(x_val)
        
        # Other metrics registered to track
        for metric in self.metrics:
            metric_value = mt.metrics(y_val, y_val_hat, type='classification')
            self.history[metric]['val'][epoch] = metric_value
    
    # If there exists regularization in the layer, it is applied only for train.
    if 'reg' in list(self.cost.keys()):
        Weights = [self.layers[layer]['W'] for layer in self.layers]
        cost_train += rg._l1_l2_EN(Weights,
                                   type=self.cost['reg']['type'],
                                   lmbda=self.cost['reg']['lmbda'],
                                   ratio=self.cost['reg']['ratio'])/x_train.shape[0]

    # Update cost value to history
    cost_value = cost_train.astype(np.float32).round(decimals=4)
    self.history[self.cost['function']]['train'][epoch] = cost_value
    
    # Any other metrics registered to track
    for metric in self.metrics:
        metric_value = mt.metrics(y_train, y_train_hat, type='classification')
        self.history[metric]['train'][epoch] = metric_value
    
    # Verbosity
    if verbosity == 3:

        print('\n- epoch:', "%3i" % epoch, '\n --------------------------------------- ', 
              '\n- cost_train:', "%.4f" % self.history[self.cost['function']]['train'][epoch],
              '- cost_val:', "%.4f" % self.history[self.cost['function']]['val'][epoch])

        if self.metrics:
            for metric in self.metrics:

                print('- ' + metric + '_train' + ': ' + 
                        "%.4f" % self.history[metric]['train'][epoch][metric],
                        '- ' + metric + '_val' + ': ' +
                        "%.4f" % self.history[metric]['val'][epoch][metric])                           

    elif verbosity == 2:
        print('coming soon: 2')
    
    elif verbosity == 1:
        print('coming soon: 1')

    # -- Backward pass
    grads = __backward_propagate(self, memory_train, y_train)

    return grads




"""
# -- --------------------------------------------------------------------------------------------------- -- #
# -- Project: lucidmode                                                                                  -- #
# -- Description: A Lightweight Framework with Transparent and Interpretable Machine Learning Models     -- #
# -- models.py: python script with Machine Learning Models                                               -- #
# -- Author: IFFranciscoME - if.francisco.me@gmail.com                                                   -- #
# -- license: GPL-3.0 License                                                                            -- #
# -- Repository: https://github.com/lucidmode/lucidmode                                                  -- #
# -- --------------------------------------------------------------------------------------------------- -- #
"""

# -- Load other scripts
import lucidmode.propagate as prop
import lucidmode.regularization as reg

# -- Load libraries for script
import numpy as np


# ------------------------------------------------------------------------------------------------------ -- #
# ------------------------------------------------------------------- FEEDFORWARD MULTILAYER PERECEPTRON -- #
# ------------------------------------------------------------------------------------------------------ -- #


class NeuralNet:

    """
    Artificial Neural Network (Feedforward multilayer pereceptron with backpropagation)

    Topology characteristics
    ------------------------

    hidden_l: Number of neurons per hidden layer (list of int, with length of l_hidden)
    hidden_a: Activation of hidden layers (list of str, with length l_hidden)   
    output_n: Number of neurons in output layer (int)
    output_a: Activation of output layer (str)
   
    Other characteristics
    ---------------------

    Layer transformations:
        - linear
        - (pending) convolution
   
    Activation functions:
        For hidden layers: 
            - Sigmoid, Tanh, ReLu (pending)
        For output layer:
            - Softmax, Linear (pending), Sigmoid (pending)
    
    Methods
    -------
    
    Weights Initialization:
        - Xavier normal, Xavier uniform, common uniform, according to [1]
        - He, according to [2]
        - (pending) Save/Load from object.
    
    Training Schemes:
        - Gradient Descent (Use all data)
            - train: use all the data on each epoch
            - validation: use all the data on each epoch
            - (pending) note for FTS: None of particular importance

        - Stochastic Gradient Descent (use 1 sample)
            - train: use 1 saample at a time and iterate through all of them on each epoch
            - validation: use all the data when train finishes, do that on each epoch 
            - (pending) note for FTS: Do not shuffle data

        - Mini-Batch Gradient Descent (use N samples)
            - train: use a partition or subset of the whole data
            - validation: use a partition or subset of the whole data (same as train)
            - (pending) note for FTS: Do not shuffle data

        - (pending) Adapting Learning
            - Momentum
            - Nesterov
        
        - (pending) Levenberg-Marquardt Algorithm

    Regularization:
        - Types: l1, l2, elasticnet, dropout
        
        - In Cost Function:
            - Weights values of all layers (l1, l2, elasticnet)
        
        - In layers
            - (pending) Weights gradient values (l1, l2, elasticnet)
            - (pending) Bias gradient values (l1, l2, elasticnet)
            - (pending) Neurons activation (dropout)

    Cost Functions: 
        - For classification: 
            - Binary Cross-Entropy 
            - Multiclass Cross-Entropy

        - For regression:
            - Mean Squared Error
    
    Execution Tools:
        - (pending) Preprocessing input data: Scale, Standard, Robust Standard.
        - (pending) Callback for termination on NaN (cost functions divergence).
        - (pending) Callback for early stopping on a metric value difference between Train-Validation sets.
        - (pending) Save weights to external object/file.
        - (pending) Load weights from external object/file.

    Visualization/Interpretation Tools: 
        - (pending) Weight values per layer (Colored bar for each neuron, separation of all layers).
        - (pending) CostFunction (train-val) evolution (two lines plot with two y-axis).
        - (pending) Convolution operation between layers.
    
    General Methods List:
        - public: fit, predict
    
    Special:
    --------

    "Ubuntu does not mean that people should not address themselves, the question, therefore is, are you
     going to do so in order to enable the community around you?", Nelson Mandela, 2006. Recorded in a 
     video made previously to the launch of Ubuntu linux distribution.

    - (pending) ubuntu_fit: Distributed learning using parallel processing among mini-batches of data 
                            selected by its value on an information divergence matrix.

    - (pending) ubuntu_predict: Voting system (classification) or Average system (regression).

    """

    # -------------------------------------------------------------------------------- CLASS CONSTRUCTOR -- #
    # -------------------------------------------------------------------------------------------------- -- #

    def __init__(self, hidden_l, hidden_a, output_n, output_a, cost=None, 
                 hidden_r=None, output_r=None, optimizer=None):

        """
        ANN Class constructor
        
        Parameters
        ----------

        hidden_l: list (of int)
            Number of neurons per hidden layer

        hidden_a: list (list of str, with length hidden_l)
            Activation of hidden layers

        output_n: int
            Number of neurons in output layer

        output_a: str
            Activation of output layer (str)
        
        hidden_r / output_r: list (of str, of size l_hidden)
            list with each pre-layer weights and biases regularization criteria, options are:

                'l1': Lasso regularization |b|
                'l2': Ridge regularization |b|^2
                'elasticnet': C(L1 - L2)
                'dropout': Randomly (uniform) select N neurons in layer and turn its weight to 0
                   
        cost: str
            cost information for model.
            'function': 'binary-logloss', 'multi-logloss', 'mse'
            'reg': {'type': ['l1', 'l2', 'elasticnet'], 'lambda': 0.001, 'ratio': 0.01}

        init: str
            initialization of weights specified from compile method

        Returns
        -------
        
        self: Modifications on instance of class

        """
        
        # Number of neurons per hidden layer
        self.hidden_l = hidden_l

        # Activation of hidden layers
        self.hidden_a = hidden_a

        # Number of neurons in output layer
        self.output_n = output_n

        # Activation of output layer (str)
        self.output_a = output_a

        # Regularization criteria for pre-output-layer weights and biases
        self.output_r = output_r

        # Cost function definition
        self.cost = cost

        # Cost function definition
        self.optimizer = optimizer

        # Regularization criteria for pre-hidden-layer weights and biases
        self.hidden_r = hidden_r
    
    # --------------------------------------------------------------------------- WEIGHTS INITIALIZATION -- #
    # -------------------------------------------------------------------------------------------------- -- #

    def __init_weights(self, input_shape, init_layers, random_state=1):
        """
        Weight initialization
        
        Parameters
        ----------

        input_shape: int
            number of features (inputs) in the model
                
        init_layers: list (of str, with size of n_layers)
        
            list with each layer criteria for weights initialization, with options: 

                'common-uniform': Commonly used factor & uniformly distributed random weights [1]
                'xavier_uniform': Xavier factor & uniformly distributed random weights [1]
                'xavier_normal': Xavier factor & standard-normally distributed random weights [1]            
                'he-standard': Factor [2]
        
        References
        ----------
        
        [1] X. Glorot and Y. Bengio.  Understanding the difficulty oftraining deep feedforward neural   
            networks. International Conference on Artificial Intelligence and Statistics, 2010.
        
        [2] He et al. "Delving Deep into Rectifiers: Surpassing Human-Level Performance on ImageNet 
            Classification," 2015 IEEE International Conference on Computer Vision (ICCV), 2015, pp. 1026-1034, doi: 10.1109/ICCV.2015.123.

        """

        # reproducibility
        np.random.seed(random_state)

        # number of hidden layers
        layers = len(self.hidden_l)

        # hidden layers weights
        for layer in range(0, layers):

            # store the type of initialization used for each layer
            self.layers['hl_' + str(layer)]['i'] = init_layers

            # number of Neurons in layer
            nn = self.hidden_l[layer]

            # multiplication factor (depends on the activation function) according to [1]
            mf = 4 if self.layers['hl_' + str(layer)]['a'] == 'tanh' else 1

            # check input dimensions for first layer
            if layer == 0:
                n_prev = input_shape
                n_next = self.hidden_l[layer]
            
            # following layers are the same
            else:
                n_prev = self.hidden_l[layer-1]
                n_next = self.hidden_l[layer]

            # As mentioned in [1]
            if init_layers == 'common-uniform':
                
                # Boundaries according to uniform distribution common heuristic
                r = mf * np.sqrt(1/nn)
                
                # Hidden layer weights and bias
                self.layers['hl_' + str(layer)]['W'] = np.random.uniform(-r, r, size=(n_next, n_prev))
                
                # Output layer
                self.layers['ol']['W'] = np.random.uniform(-r, r, size=(self.output_n, self.hidden_l[-1]))
                
                # Bias weigths in zero
                self.layers['hl_' + str(layer)]['b'] = np.zeros((nn, 1))
                self.layers['ol']['b'] = np.zeros((self.output_n, 1))

            # According to eq:16 in [1]
            elif init_layers == 'xavier-uniform':
                
                # Boundaries according to uniform distribution common heuristic
                r = mf * np.sqrt(6/(n_prev + n_next))
                
                # Hidden layer weights and bias
                self.layers['hl_' + str(layer)]['W'] = np.random.uniform(-r, r, size=(n_next, n_prev))

                # Output layer
                self.layers['ol']['W'] = np.random.uniform(-r, r, size=(self.output_n, self.hidden_l[-1]))
                
                # Bias weigths in zero
                self.layers['hl_' + str(layer)]['b'] = np.zeros((nn, 1))
                self.layers['ol']['b'] = np.zeros((self.output_n, 1))

            # A variation of the previous, according to [1]
            elif init_layers == 'xavier-standard':
                
                # Multiplying factor (paper version)
                r = mf * np.sqrt(2/(n_prev + n_next))
                
                # Hidden layer weights and biasW
                self.layers['hl_' + str(layer)]['W'] = np.random.randn(n_next, n_prev) * r
                
                # Output layer
                self.layers['ol']['W'] = np.random.randn(self.output_n, self.hidden_l[-1]) * r
                
                # Bias weigths in zero
                self.layers['hl_' + str(layer)]['b'] = np.zeros((nn, 1))
                self.layers['ol']['b'] = np.zeros((self.output_n, 1))

           # According to [2]
            elif init_layers == 'he-standard':
                
                # Multiplying factor
                r = mf * np.sqrt(2/(n_prev + n_next))
                
                # Hidden layer weights and bias
                self.layers['hl_' + str(layer)]['W'] = np.random.randn(n_next, n_prev) * r
                
                # Output layer
                self.layers['ol']['W'] = np.random.randn(self.output_n, self.hidden_l[-1]) * r

                # Bias weigths in zero
                self.layers['hl_' + str(layer)]['b'] = np.zeros((nn, 1))
                self.layers['ol']['b'] = np.zeros((self.output_n, 1))

            else: 
                print('Raise Error')

    
    # --------------------------------------------------------------------------------- LAYERS FORMATION -- #
    # -------------------------------------------------------------------------------------------------- -- #


    def formation(self, cost=None, optimizer=None, init=None, metrics=None):
        """
        Neural Network Model Formation.        
        
        Parameters
        ----------
        
        self: Instance of class

        cost: 
            cost_f: Cost function
            cost_r: Cost regularization
        
        optimizer: 
            type: Name of method for optimization
            params: parameters according to method
        
        init:
            weight initialization
        
        metrics: 
            metrics to monitor training


        Returns
        -------
        
        self: Modifications on instance of class

        """

        # Hidden layers
        self.layers = {'hl_' + str(layer): {'W': {}, 'b':{}, 'a': {}, 'r':self.hidden_r[layer]}
                       for layer in range(0, len(self.hidden_l))}

        # Output layer
        self.layers.update({'ol': {'W': {}, 'b': {}, 'a': self.output_a, 'r': self.output_r}})

        # iterative layer formation loop
        for layer in range(0, len(self.hidden_l)):

            # layer neurons composition
            self.layers['hl_' + str(layer)]['W'] = None

            # layer biases
            self.layers['hl_' + str(layer)]['b'] = None

            # layer activation
            # if only 1 activation function was provided, use it for all hidden layers
            act = self.hidden_a[0] if len(self.hidden_a) == 1 else self.hidden_a[layer]
            self.layers['hl_' + str(layer)]['a'] = act

            # layer regularization
            self.layers['hl_' + str(layer)]['r'] = self.hidden_r[layer]

            # layer weights initialization
            self.layers['hl_' + str(layer)]['i'] = ''
        
        # Weights initialization
        self.__init_weights(input_shape=init['input_shape'], init_layers=init['init_layers'])

        # Cost (function and regularization definition)
        self.cost = cost
        
        # Metrics to track progress on learning
        self.metrics = metrics

        # Optimizer
        self.optimizer = optimizer
        
        # Store evolution of cost and other metrics across epochs
        history = {self.cost['function']: {'train': {}, 'val': {}}}
        history.update({metric: {'train': {}, 'val': {}} for metric in self.metrics})
        self.history = history

    # ------------------------------------------------------------------ FIT MODEL PARAMETERS (LEARNING) -- #
    # -------------------------------------------------------------------------------------------------- -- #


    def fit(self, x_train, y_train, x_val=None, y_val=None, epochs=10, alpha=0.1, verbosity=3,
            random_state=1, callback=None, randomize=False):
        
        """
        Train model according to specified parameters

        Parameters
        ----------

        x_train: np.array / pd.Series
            Features data with nxm dimensions, n = observations, m = features
        
        y_train: np.array / pd.Series
            Target variable data, dimensions of: nx1 por binary classification and nxm for multi-class
        
        x_val: np.array / pd.Series
            Same as x_train but with data considered as validation

        y_val: np.array / pd.Series
            Same as y_train but with data considered as validation

        epochs: int
            Epochs to iterate the model training
        
        alpha: float
            Learning rate for Gradient Descent
        
        cost_f: str
            Cost function, options are according to functions.cost

        verbosity: int
            level of verbosity to show progress
            3: cost train and cost val at every epoch
        
        callback: dict
            whether there is a stopping criteria or action

            {'earlyStopping': {'metric': 'acc', 'threshold': 0.80}}
        
        Returns
        -------

        history: dict
            with dynamic keys and iterated values of selected metrics

        # binary output
        # y_train = data['y'].astype(np.int)

        """ 

        import learning.execution as ex

        # Store callbacks in class
        self.callbacks = callback

        # ------------------------------------------------------------------------------ TRAINING EPOCHS -- #
        
        epoch_count = epochs
        epoch_for = 0

        while epoch_count > 0:

            epoch_count -= 1
            epoch_for += 1

            # reproducibility
            np.random.seed(random_state)
            m_train = x_train.shape[0]

            # -- Stochastic Gradient Descent
            if self.optimizer['type'] == 'SGD':

                # get value for batch size
                batch_size = self.optimizer['params']['batch_size']
                
                # if batch size is 0 then use all of samples
                batch_size = m_train if batch_size == 0 else batch_size

                # No random sampling by default
                s_x_train = x_train
                s_y_train = y_train

                if randomize:
                    # randomize samples
                    perm_train = list(np.random.permutation(m_train))
                    s_x_train = x_train[perm_train, :]
                    s_y_train = y_train[perm_train]

                # number of batches
                n_train = np.trunc(m_train / batch_size).astype(int)

                # iterate over all batches
                for k in range(n_train):
                    
                    batch_x = s_x_train[k*batch_size : (k + 1)*batch_size, :]
                    batch_y = s_y_train[k*batch_size : (k + 1)*batch_size]

                    grads = prop._forward_backward(self, batch_x, batch_y, x_val=x_val, y_val=y_val,
                                                   epoch=epoch_for)

                    # ------------------------------------------------------------------- CALLBACK CHECK -- #
                    if ex.callback_es(self, epoch_for) == 'triggered':
                        return 'callback triggered'
                    
                # If remaining batch is left, iterate over it
                if m_train % batch_size != 0:
                    
                    num_last_batch = m_train - (batch_size * n_train)
                    batch_x = s_x_train[m_train - num_last_batch:m_train, :]
                    batch_y = s_y_train[m_train - num_last_batch:m_train]

                    grads = prop._forward_backward(self, batch_x, batch_y, x_val=x_val, y_val=y_val,
                                                   epoch=epoch_for)

                    # ------------------------------------------------------------------- CALLBACK CHECK -- #
                    if ex.callback_es(self, epoch_for) == 'triggered':
                        return 'callback triggered'

                # Update all layers weights and biases
                for l in range(0, len(self.hidden_l) + 1):

                    # Model Elements
                    layer  = list(self.layers.keys())[l]               
                    dW = grads['dW_' + str(l + 1)]
                    W = self.layers[layer]['W']
                    db = grads['db_' + str(l + 1)]
                    b = self.layers[layer]['b']
                    
                    # If the layer has regularization criteria
                    if self.layers[layer]['r']:
                        r_t = self.layers[layer]['r']['type']
                        r_l = self.layers[layer]['r']['lmbda']
                        r_r = self.layers[layer]['r']['ratio']
                        regW = reg._l1_l2_EN([W], type=r_t, lmbda=r_l, ratio=r_r)/x_train.shape[0]
                        regb = reg._l1_l2_EN([b], type=r_t, lmbda=r_l, ratio=r_r)/x_train.shape[0]
                    
                    # No regularization
                    else:
                        regW, regb = 0

                    # Gradient updating
                    self.layers[layer]['W'] = W - (self.optimizer['params']['learning_rate'] * dW) + regW
                    self.layers[layer]['b'] = b - (self.optimizer['params']['learning_rate'] * db) + regb
                                           
                # ----------------------------------------------------------------------- CALLBACK CHECK -- #
                if ex.callback_es(self, epoch_for) == 'triggered':
                    return 'callback triggered'

            # -- levenberg-Marquardt Algorithm
            elif self.optimizer['type'] == 'LMA':
                
                return 'coming soon'


    # ------------------------------------------------------------------------------- PREDICT WITH MODEL -- #
    # -------------------------------------------------------------------------------------------------- -- #


    def predict(self, x_train):
        """

        PREDICT depends of the activation function and number of outpus in output layer

        """
        
        # -- SINGLE-CLASS
        if self.output_n == 1:            
            # tested only with sigmoid output
            memory = prop._forward_propagate(self, x_train)
            p = memory['A_' + str(len(self.hidden_l) + 2)]
            thr = 0.5
            indx = p > thr
            p[indx] = 1
            p[~indx] = 0

        # -- MULTI-CLASS 
        else:
            memory = prop._forward_propagate(self, x_train)
            p = np.argmax(memory['A_' + str(len(self.hidden_l) + 2)], axis=1)

        return p




"""
# -- --------------------------------------------------------------------------------------------------- -- #
# -- Project: lucidmode                                                                                  -- #
# -- Description: A Lightweight Framework with Transparent and Interpretable Machine Learning Models     -- #
# -- convolution.py: python script with convolution functions                                            -- #
# -- Author: IFFranciscoME - if.francisco.me@gmail.com                                                   -- #
# -- license: GPL-3.0 License                                                                            -- #
# -- Repository: https://github.com/lucidmode/lucidmode                                                  -- #
# -- --------------------------------------------------------------------------------------------------- -- #
"""

# -- Load libraries for script
import numpy as np
from scipy import signal
from matplotlib import pyplot as plt

# -- Load other scripts

# ----------------------------------------------------------------------------------- CONVOLUTION LAYERS -- #

"""

Work in progress for 1d convolution layer

References
----------

https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.fftconvolve.html
https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.windows.gaussian.html

"""

# The option would be to choose

# 1) Outputs of past layer are inputs for conv layer, 
# 2) Define type of kernel: Gaussian noise
# 3) Specify memory of convolution: from 1 to batch_size
# 4) Specify kernel parameters: 1) Distribution Variance
# 5) Scale output of convolution: between 0 to 1
# 6) To apply an activation function after convolution: sigmoid, tanh, relu

# Challenge 1) How to compute the gradient of J with respect to the output of 1d Conv layer?
# Since output of convolution goes through an activation function, forward and backward are according to 
# such activation

np.random.seed(123) 

# Previous layer example
y = np.cumsum(np.random.randn(10))

# Previous layer number of inputs
prev_n = 10

# Memory to convolve
param_m = 1

# -- Kernel Parameters {'k_0': Type of kernel, 'k_1': Variance}
param_k_0 = 'gaussian'
param_k_1 = 1.5

# -- Fast Fourier Convolution Operation
z = signal.windows.gaussian(param_m, param_k_1)
y_convolved = signal.fftconvolve(y, z, mode='same')

# -- Scaling, Dimensions and Shape
y_convolved = y_convolved/np.max(abs(y_convolved))
y = y/np.max(abs(y))

y_convolved = np.matrix(y_convolved).T

plt.plot(y, label='original')
plt.legend()
plt.title('Original TS - Gaussian Random Walk')
 
plt.plot(y_convolved, label='convolved')
plt.legend()
plt.title('Convolved TS')

plt.show()
